# -*- coding: utf-8 -*-
""" Base class for resources based on DB models. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from logging import getLogger

# 3rd party imports
from jsonschema import validate, ValidationError
from serafin import Fieldspec, serialize
from six import iteritems

# local imports
from .resource import RestResource
from .util import iter_public_props


L = getLogger(__name__)


class ModelResource(RestResource):
    """ Base class for resources based on DB models. """

    model = None
    spec = Fieldspec('*')
    schema = {}
    read_only = []

    class AlreadyExists(RuntimeError):
        """ Raised when an object already exists. """
        pass

    class ValidationError(RuntimeError):
        """ Raised by .validate() if it fails. """
        def __init__(self, jsonschema_error):
            self.detail = jsonschema_error
            super(ModelResource.ValidationError, self).__init__(
                str(jsonschema_error)
            )

    def validate(self, data, schema=None):
        """ Validate the *data* according to the given *schema*.

        :param Dict[str, Any] data:
            A dictionary of data. Probably coming from the user in some way.
        :param Dict[str, Any] schema:
            JSONSchema describing the required data structure.
        :raises ModelResource.ValidationError:
            If the validation fails. No value is returned.
        """
        try:
            validate(data, schema or self.schema)
        except ValidationError as ex:
            raise ModelResource.ValidationError(ex)

    def serialize(self, item_or_items):
        """ Serialize an item or items into a dict.

        This will just call serafin.serialize using the model spec (defined
        class wide in the model - '*' by default).

        :param item_or_items:
        :return Dict[Any, Any]:
            A dict with python native content. Can be easily dumped to any
            format like JSON or YAML.
        """
        return serialize(item_or_items, self.spec)

    def deserialize(self, data):
        """ Convert JSON data into model field types.

        The value returned by this function can be used directly to create new
        item and update existing ones.
        """
        return {n: self.get_field_value(n, v) for n, v in iteritems(data)}

    def get_field_value(self, name, value):
        """ Coerce value to a model field compatible representation. """
        return value

    @property
    def public_props(self):
        """ All public properties on the resource model. """
        if self._public_props is None:
            self._public_props = [
                name for name, _ in iter_public_props(self.model)
            ]
        return self._public_props

    def create_item(self, values):
        """ Create new model item. """
        raise NotImplementedError("Must implement .create_item()")

    def update_item(self, request, values):
        """ Update existing model item. """
        raise NotImplementedError("Must implement .update_item()")

    def delete_item(self, item):
        """ Update existing model item. """
        raise NotImplementedError("Must implement .delete_item()")

    def dbquery(self, request, filters):
        """ Return a model query with the given filters.

        The query can be further customised like any ndb query.

        :return google.appengine.ext.ndb.Query:
            The query with the given filters already applied.
        """
        raise NotImplementedError("Must implement .dbquery()")

    def get_requested(self, request):
        """ Get requested item. """
        raise NotImplementedError("Must implement .get_requested()")

    def rest_query(self, request, params):
        """ Query existing records as a list. """
        try:
            fields = params.pop('_fields', '*')

            filters = self.deserialize(params)
            items = self.dbquery(request, filters)

            spec = Fieldspec(self.spec).restrict(Fieldspec(fields))
            ret = serialize(items, spec)
            return 200, ret

        except NotImplementedError:
            return 405, {'detail': 'Method not allowed'}

    def rest_create(self, request, data):
        """ Create a new record. """
        try:
            self.validate(data, self.schema)

            values = self.deserialize(data)
            item = self.create_item(values)

            return serialize(item, self.spec)

        except ModelResource.ValidationError as ex:
            return 400, {'detail': str(ex)}

        except ModelResource.AlreadyExists:
            return 400, {'detail': 'Already exists'}

        except NotImplementedError:
            return 405, {'detail': 'Method not allowed'}

    def rest_get(self, request, params):
        """ Get one record with the given id. """
        try:
            fields = Fieldspec(params.get('_fields', '*'))

            spec = Fieldspec(self.spec).restrict(fields)
            item = self.get_requested(request)

            if item is not None:
                return 200, serialize(item, spec)
            else:
                return 404, {'detail': "Not found"}

        except NotImplementedError:
            return 405, {'detail': 'Method not allowed'}

    def rest_update(self, request, data):
        """ Update existing item. """
        schema = {}
        schema.update(self.schema)

        if 'required' in schema:
            del schema['required']

        try:
            self.validate(data, schema)

            values = self.deserialize(data)
            read_only = frozenset(self.read_only) | frozenset(self.public_props)
            for name in read_only:
                values.pop(name, None)

            item = self.update_item(request, values)

            if item is not None:
                return 200, serialize(item, self.spec)
            else:
                return 404, {'detail': "Not Found"}

        except ModelResource.ValidationError as ex:
            return 400, {'detail': str(ex)}

        except NotImplementedError:
            return 405, {'detail': 'Method not allowed'}

    def rest_delete(self, request):
        """ DELETE detail. """
        try:
            item = self.get_requested(request)

            if item is None:
                return 404, {'detail': 'Item does not exist'}

            self.delete_item(item)

            return 204, {}

        except NotImplementedError:
            return 405, {'detail': 'Method not allowed'}



# -*- coding: utf-8 -*-
"""
Base class for resources using Google AppEngine ndb as storage.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from logging import getLogger

# 3rd party imports
from serafin import Fieldspec
from six import iteritems

# local imports
from restible.core.model import ModelResource


L = getLogger(__name__)


class NdbResource(ModelResource):
    """ Base class for ndb based resources.

    This provides a basic implementation that can be used out of the box. It
    provides no authentication/authorization.
    """
    name = None
    model = None
    spec = Fieldspec('*')
    schema = {}

    def create_item(self, values):
        item = self.model(**values)
        item.put()

        return item

    def update_item(self, request, values):
        item = self.get_requested(request)

        if item is None:
            return None

        for name, value in iteritems(values):
            if name not in ('id',):
                try:
                    setattr(item, name, value)
                except AttributeError:
                    L.exception("Failed to set attribute '{}'".format(name))
                    raise

        item.put()

        return item

    def delete_item(self, item):
        item.delete()

    def dbquery(self, request, filters):
        """ Return a model query with the given filters.

        The query can be further customised like any ndb query.

        :return google.appengine.ext.ndb.Query:
            The query with the given filters already applied.
        """
        filters = [getattr(self.model, n) == v for n, v in iteritems(filters)]
        return self.model.query().filter(*filters)

    def get_requested(self, request):
        pk = self.get_pk(request)
        return self.model.get_by_id(pk)



# -*- coding: utf-8 -*-
"""
webapp2 implementation for the `RestEndpoint`.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import json

# 3rd party imports
import webapp2
from six import iteritems

# local imports
from restible import RestEndpoint, RawResponse


class Webapp2Endpoint(webapp2.RequestHandler, RestEndpoint):
    """ Endpoint implementation to use in webapp2/AppEngine projects. """
    def __init__(self, *args, **kw):
        webapp2.RequestHandler.__init__(self, *args, **kw)
        RestEndpoint.__init__(self, self.resource)

    @classmethod
    def extract_request_data(cls, request):
        return json.loads(request.body)

    def response_from_result(self, result):
        """ Generate webapp2 response from  RestResult.

        :param RestResult result:
            RestResult instance with the API call result.
        """
        if isinstance(result, RawResponse):
            pass

        for name, value in iteritems(result.headers):
            self.response.headers[name] = value

        self.response.set_status(result.status)
        self.response.out.write(json.dumps(result.data))

    def dispatch(self):
        """ Override webapp2 dispatcher. """
        request = self.request
        request.rest_keys = request.route_kwargs

        result = self.call_rest_handler(request.method, request)
        return self.response_from_result(result)



# -*- coding: utf-8 -*-
"""
Simple REST API implementation.

Each endpoint has it's resource class that represents all the operations
possible in that endpoint. The resource implementation should be framework
agnostic so it can be used both on AppEngine and regular django (possibly other
frameworks as well..
"""
from __future__ import absolute_import, unicode_literals

# package interface
from .core.endpoint import RestEndpoint, RawResponse
from .core.model import ModelResource
from .core.resource import RestResource
from .core.routing import api_action
from .core.routing import api_route

__version__ = '0.3.1'

__all__ = [
    'api_action',
    'api_route',
    'ModelResource',
    'RawResponse',
    'RestEndpoint',
    'RestResource',
]



# -*- coding: utf-8 -*-
"""
Base class for django based REST resources
"""
from __future__ import absolute_import, unicode_literals

# 3rd party imports
from django.db.utils import IntegrityError

# local imports
from restible.core.model import ModelResource


class DjangoResource(ModelResource):
    """ Base class for django based REST resources. """
    def create_item(self, values):
        try:
            return self.model.objects.create(**values)
        except IntegrityError as ex:
            raise ModelResource.AlreadyExists(str(ex))

    def update_item(self, request, values):
        pk = self.get_pk(request)
        return self.model.objects.filter(pk=pk).update(**values)

    def delete_item(self, item):
        item.delete()

    def dbquery(self, request, filters):
        return self.model.objects.filter(**filters)

    def get_requested(self, request):
        """ Get requested item. """
        pk = self.get_pk(request)
        try:
            return self.dbquery(request, {}).get(pk=pk)
        except self.model.DoesNotExist:
            return None



# -*- coding: utf-8 -*-
"""
Helpers for setting up mapping between URLs and resources.
"""
from __future__ import absolute_import, unicode_literals

# 3rd party imports
import attr


@attr.s
class ActionMeta(object):
    """ Action metadata.

    When you have a action method ``Res.my_action`` you can access the metadata
    with ``api_action.get_meta(Res.my_action)``.
    """
    name = attr.ib()
    generic = attr.ib()
    protected = attr.ib()
    methods = attr.ib()


@attr.s
class RouteMeta(object):
    """ Route metadata.

    When you have a route function ``my_route`` you can access the metadata
    with ``api_route.get_meta(my_route)``.
    """
    methods = attr.ib()


class api_route(object):
    """ Decorator for defining API routs.

    API routes are endpoints that do not belong to any resource.
    """
    ROUTE_ATTR = '__api_route__'

    def __init__(self, methods=None):
        self.meta = RouteMeta(
            methods=methods or ['post']
        )

    def __call__(self, fn):
        """ Decorator. """
        setattr(fn, self.ROUTE_ATTR, self.meta)
        return fn

    @classmethod
    def get_meta(cls, route):
        """ Get route metadata. """
        return getattr(route, cls.ROUTE_ATTR, {})


class api_action(object):
    """ Decorator for defining resource actions.

    Actions are extra HTTP endpoints that can be called with POST.
    """
    ACTION_ATTR = '__api_action__'

    def __init__(self, name=None, generic=False, protected=True, methods=None):
        self.name = name
        self.generic = generic
        self.protected = protected
        self.methods = [m.lower() for m in (methods or ['post'])]

    def __call__(self, fn):
        """ Decorator. """
        setattr(fn, self.ACTION_ATTR, ActionMeta(
            name=self.name or fn.__name__,
            generic=self.generic,
            protected=self.protected,
            methods=self.methods,
        ))
        return fn

    @classmethod
    def is_action(cls, obj):
        """ Return true if obj is an API action. """
        return hasattr(obj, cls.ACTION_ATTR)

    @classmethod
    def get_meta(cls, action):
        """ Get action metadata """
        return getattr(action, cls.ACTION_ATTR, {})
